/* 
   NSConcreteString.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <config.h>

#include <ctype.h>

#include <Foundation/common.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSData.h>
#include <Foundation/NSCoder.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSException.h>
#include <Foundation/exceptions/StringExceptions.h>
#include <Foundation/exceptions/GeneralExceptions.h>

#include <Foundation/NSCharacterSet.h>
#include <Foundation/NSString.h>
#include <Foundation/NSConcreteString.h>

#include <extensions/objc-runtime.h>

#define COLLECT_STRING_CLUSTER_STATISTICS 0

#if COLLECT_STRING_CLUSTER_STATISTICS
static unsigned int NS8BitString_dealloc_count             = 0;
static unsigned int NSInline8BitString_dealloc_count       = 0;
static unsigned int NSInline8BitString_total_len           = 0;
static unsigned int NSShortInline8BitString_dealloc_count  = 0;
static unsigned int NSShortInline8BitString_total_len      = 0;
static unsigned int NSNonOwned8BitString_dealloc_count     = 0;
static unsigned int NSNonOwned8BitString_total_len         = 0;
static unsigned int NSOwned8BitString_dealloc_count        = 0;
static unsigned int NSOwned8BitString_total_len            = 0;
static unsigned int NSNonOwnedOpen8BitString_dealloc_count = 0;
static unsigned int NSNonOwnedOpen8BitString_total_len     = 0;
static unsigned int NSOwnedOpen8BitString_dealloc_count    = 0;
static unsigned int NSOwnedOpen8BitString_total_len        = 0;
static unsigned int NSRange8BitString_dealloc_count        = 0;
static unsigned int NSRange8BitString_total_len            = 0;

@implementation NSString(ClusterStatistics)

+ (void)printStatistics
{
    fprintf(stderr,
            "NSString class cluster statistics:\n"
            "  dealloc counts:\n"
            "    NS8BitString:                 %d\n"
            "      NSInline8BitString:         %d\n"
            "      NSShortInline8BitString:    %d\n"
            "      NSNonOwned8BitString:       %d\n"
            "        NSOwned8BitString:        %d\n"
            "          NSOwnedOpen8BitString:  %d\n"
            "        NSNonOwnedOpen8BitString: %d\n"
            "          NSRange8BitString:      %d\n"
            "  avg len (dealloc statistics):\n"
            "    NS8BitString:\n"
            "      NSInline8BitString:         %d\n"
            "      NSShortInline8BitString:    %d\n"
            "      NSNonOwned8BitString:       %d\n"
            "        NSOwned8BitString:        %d\n"
            "          NSOwnedOpen8BitString:  %d\n"
            "        NSNonOwnedOpen8BitString: %d\n"
            "          NSRange8BitString:      %d\n"
            ,
            NS8BitString_dealloc_count,
            NSInline8BitString_dealloc_count,
            NSShortInline8BitString_dealloc_count,
            NSNonOwned8BitString_dealloc_count,
            NSOwned8BitString_dealloc_count,
            NSOwnedOpen8BitString_dealloc_count,
            NSNonOwnedOpen8BitString_dealloc_count,
            NSRange8BitString_dealloc_count,
            NSInline8BitString_dealloc_count
              ? NSInline8BitString_total_len / NSInline8BitString_dealloc_count
              : 0,
            NSShortInline8BitString_dealloc_count
              ? NSShortInline8BitString_total_len /
                NSShortInline8BitString_dealloc_count
              : 0,
            NSNonOwned8BitString_dealloc_count
              ? NSNonOwned8BitString_total_len/NSNonOwned8BitString_dealloc_count
              : 0,
            NSOwned8BitString_dealloc_count
              ? NSOwned8BitString_total_len / NSOwned8BitString_dealloc_count
              : 0,
            NSOwnedOpen8BitString_dealloc_count
              ? NSOwnedOpen8BitString_total_len /
                NSOwnedOpen8BitString_dealloc_count
              : 0,
            NSNonOwnedOpen8BitString_dealloc_count
              ? NSNonOwnedOpen8BitString_total_len /
                NSNonOwnedOpen8BitString_dealloc_count
              : 0,
            NSRange8BitString_dealloc_count
              ? NSRange8BitString_total_len / NSRange8BitString_dealloc_count
              : 0
            );
}
- (void)printStatistics
{
    [NSString printStatistics];
}

@end
#endif /* COLLECT_STRING_CLUSTER_STATISTICS */

@implementation NS8BitString

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NS8BitString_dealloc_count++;
    [super dealloc];
}
#endif

/* Accessing characters	*/

- (void)getCharacters:(unichar*)buffer range:(NSRange)aRange
{
    if (aRange.location + aRange.length > [self cStringLength])
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, [self cStringLength]]);
    
    {
	int i = 0;
	unsigned char *bytes = [self __compact8BitBytes];

	for (i = 0; i < aRange.length; i++)
	    buffer[i] = bytes[i];
    }
}

/* Dividing strings */

- (NSString *)substringWithRange:(NSRange)aRange
{
    [self subclassResponsibility:_cmd];
    return nil;
}

/* Finding characters and substrings */

- (NSRange)rangeOfCharacterFromSet:(NSCharacterSet*)aSet
  options:(unsigned int)mask range:(NSRange)aRange
{
    // ENCODINGS - this code applies to the system's default encoding
    int i = 0;
    NSRange range = {0, 0};
    IMP imp = [aSet methodForSelector:@selector(characterIsMember:)];
    unsigned char *bytes = [self __compact8BitBytes];

    if (aRange.location + aRange.length > [self cStringLength])
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, [self cStringLength]]);

    if (mask & NSBackwardsSearch) {
	for (i = aRange.length - 1; i >= 0; i--) {
	    unichar c = bytes[i];
	    
	    if ((*imp)(aSet, @selector(characterIsMember:), c) ||
		((mask & NSCaseInsensitiveSearch) && 
		 ((islower(c) &&
		  (*imp)(aSet, @selector(characterIsMember:), toupper(c))) ||
		 (isupper(c) &&
		  (*imp)(aSet, @selector(characterIsMember:), tolower(c))))
		 )) {
		    range.location = aRange.location + i;
		    range.length = 1;
		    return range;
		}
	}
    } 
    else {
	for (i = 0; i < aRange.length; i++) {
	    unichar c = bytes[i];

	    if ((*imp)(aSet, @selector(characterIsMember:), c) ||
		((mask & NSCaseInsensitiveSearch) && 
		 ((islower(c) &&
		  (*imp)(aSet, @selector(characterIsMember:), toupper(c))) ||
		 (isupper(c) &&
		  (*imp)(aSet, @selector(characterIsMember:), tolower(c))))
		 )) {
		    range.location = aRange.location + i;
		    range.length = 1;
		    return range;
		}
	}
    }
    
    return range;
}

- (NSRange)rangeOfString:(NSString*)aString
  options:(unsigned int)mask range:(NSRange)aRange
{
    // ENCODINGS - this code applies to the system's default encoding
    NSRange       range;
    unsigned char *mbytes;
    unsigned char *abytes;
    int           a;
    
    if (![aString isKindOfClass:[NS8BitString class]] &&
	![aString isKindOfClass:[NSMutable8BitString class]])
	    return [super rangeOfString:aString options:mask range:aRange];
    
    if ((aRange.location + aRange.length) > [self cStringLength]) {
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, [self cStringLength]]);
    }

    mbytes = [self __compact8BitBytes] + aRange.location;
    abytes = [(id)aString __compact8BitBytes];
    a = [aString cStringLength];
    
    if ((a == 0) || (aRange.length < a))
        return NSMakeRange(0, 0);
    
    if (mask & NSAnchoredSearch)  {
	range.location = aRange.location + 
	    ((mask & NSBackwardsSearch) ? aRange.length - a : 0);
	range.length = a;
	
	if ([self compare:aString options:mask range:range] == NSOrderedSame)
	    return range;
	else
	    return NSMakeRange(0,0);
    }
    
    if (mask & NSBackwardsSearch) {	
	if (mask & NSCaseInsensitiveSearch) {
	    /* Backward case insensitive */
	    unsigned char cf;
            int n;

            cf = islower(abytes[0]) ? toupper(abytes[0]) : abytes[0];
            
	    for (n = aRange.length-a; n >= 0; n--) {
		unsigned char cm =
                    islower(mbytes[n]) ? toupper(mbytes[n]) : mbytes[n];
		unsigned char ca = cf;
                int i;
                
		if (cm != ca)
		    continue;
		for (i = 1; i < a; i++) {
		    cm = islower(mbytes[n+i]) ? 
			toupper(mbytes[n+i]) : mbytes[n+i];
		    ca = islower(abytes[i]) ? toupper(abytes[i]) : abytes[i];
		    if (cm != ca)
			break;
		}
		if (i == a) {
		    range.location = aRange.location + n;
		    range.length = a;
		    return range;
		}
	    }
	}
	else {
	    /* Backward case sensitive */
            int n;
	    for (n = (aRange.length - a); n >= 0; n--) {
                int i;
                
		if (mbytes[n] != abytes[0])
		    continue;
		for (i = 1; i < a; i++)
		    if (mbytes[n+i] != abytes[i])
			break;
		if (i == a) {
		    range.location = aRange.location + n;
		    range.length = a;
		    return range;
		}
	    }
	}
    }
    else {
	if (mask & NSCaseInsensitiveSearch) {
	    /* Forward case insensitive */
            int n;
	    unsigned char cf;

            cf = islower(abytes[0]) ? toupper(abytes[0]) : abytes[0];

	    for (n = 0; n + a <= aRange.length; n++) {
		unsigned char cm, ca;
                int i;
		
                cm = islower(mbytes[n]) ? toupper(mbytes[n]) : mbytes[n];
                ca = cf;
                
		if (cm != ca)
		    continue;
		for (i = 1; i < a; i++) {
		    cm = islower(mbytes[n+i]) ? 
			toupper(mbytes[n+i]) : mbytes[n+i];
		    ca = islower(abytes[i]) ? toupper(abytes[i]) : abytes[i];
		    if (cm != ca)
			break;
		}
		if (i == a) {
		    range.location = aRange.location + n;
		    range.length = a;
		    return range;
		}
	    }
	}
	else {
	    /* Forward case sensitive */
            int n;
            
	    for (n = 0; (n + a) <= aRange.length; n++) {
                int i;
                
		if (mbytes[n] != abytes[0])
		    continue;
		for (i = 1; i < a; i++)
		    if (mbytes[n+i] != abytes[i])
			break;
		if (i == a) {
		    range.location = aRange.location + n;
		    range.length   = a;
		    return range;
		}
	    }
	}
    }
    
    range.location = range.length = 0;
    return range;
}

- (NSComparisonResult)compare:(NSString*)aString
  options:(unsigned int)mask range:(NSRange)aRange
{
    // ENCODINGS - this code applies to the system's default encoding
    unsigned char* mbytes;
    unsigned char* abytes;
    int i, n, a;

    if (![aString isKindOfClass:[NS8BitString class]] &&
	![aString isKindOfClass:[NSMutable8BitString class]])
	    return [super compare:aString options:mask range:aRange];
    
    if (aRange.location + aRange.length > [self cStringLength])
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, [self cStringLength]]);

    mbytes = [self __compact8BitBytes] + aRange.location;
    abytes = [(id)aString __compact8BitBytes];
    
    a = [aString cStringLength];
    n = MIN(a, aRange.length);
    
    if (mask & NSCaseInsensitiveSearch) {
	for (i = 0; i < n; i++) {
	    unsigned char cm = islower(mbytes[i]) ? toupper(mbytes[i]):mbytes[i];
	    unsigned char ca = islower(abytes[i]) ? toupper(abytes[i]):abytes[i];

	    if (cm < ca)
		return NSOrderedAscending;
	    if (cm > ca)
		return NSOrderedDescending;
	}
    }
    else {
	for (i = 0; i < n; i++) {
	    if (mbytes[i] < abytes[i])
		return NSOrderedAscending;
	    if (mbytes[i] > abytes[i])
		return NSOrderedDescending;
	}
    }
    
    if (aRange.length < a)
	return NSOrderedAscending;
    if (aRange.length > a)
	return NSOrderedDescending;

    return NSOrderedSame;
}

- (unsigned)hash
{
    unsigned char* bytes = [self __compact8BitBytes];
    unsigned hash = 0, hash2;
    int i, n = [self cStringLength];

    for(i=0; i < n; i++) {
        hash <<= 4;
	// UNICODE - must use a for independent of composed characters
        hash += bytes[i];
        if((hash2 = hash & 0xf0000000))
            hash ^= (hash2 >> 24) ^ hash2;
    }
    
    return hash;
}

/* Getting a shared prefix */

- (NSString *)commonPrefixWithString:(NSString*)aString
  options:(unsigned int)mask
{
    // ENCODINGS - this code applies to the system's default encoding
    NSRange range = {0, 0};
    unsigned char *mbytes;
    unsigned char *abytes;
    int           mLen, aLen, i;
    
    if (![aString isKindOfClass:[NS8BitString class]] &&
	![aString isKindOfClass:[NSMutable8BitString class]])
	    return [super commonPrefixWithString:aString options:mask];
    
    mLen   = [self cStringLength];
    aLen   = [aString length];
    mbytes = [self __compact8BitBytes];
    abytes = [(NS8BitString *)aString __compact8BitBytes];

    for (i = 0; (i < mLen) && (i < aLen); i++) {
	unsigned char c1 = mbytes[i];
	unsigned char c2 = abytes[i];
        
        if (mask & NSCaseInsensitiveSearch) {
            c1 = tolower(c1);
            c2 = tolower(c2);
        }
        if (c1 != c2)
            break;
    }
    
    range.length = i;
    return [self substringWithRange:range];
}

/* Changing case */

- (NSString *)capitalizedString
{
    // ENCODINGS - this code applies to the system's default encoding
    int            i;
    BOOL           f      = YES;
    int            length = [self cStringLength];
    unsigned char* bytes  = [self __compact8BitBytes];
    unsigned char* chars  = MallocAtomic(sizeof(unichar)*(length+1));

    for (i = 0; i < length; i++) {
	unsigned char c = bytes[i];
	
	if (isspace(c))
	    f = YES;
	
	if (f) {
	    chars[i] = islower(c) ? toupper(c) : c;
	    f = NO;
	}
	else
	    chars[i] = isupper(c) ? tolower(c) : c;
    }
    chars[i] = 0;
    
    return AUTORELEASE([[NSOwned8BitString alloc]
                           initWithCString:chars length:length copy:NO]);
}

- (NSString *)lowercaseString
{
    // ENCODINGS - this code applies to the system's default encoding
    int i;
    int length = [self cStringLength];
    unsigned char *bytes = [self __compact8BitBytes];
    unsigned char *chars = MallocAtomic(sizeof(unichar)*(length+1));

    for (i = 0; i < length; i++) {
	register unsigned char c = bytes[i];
	chars[i] = isupper(c) ? tolower(c) : c;
    }
    chars[i] = 0;

    return AUTORELEASE([[NSOwned8BitString alloc]
                           initWithCString:chars length:length copy:NO]);
}

- (NSString *)uppercaseString
{
    // ENCODINGS - this code applies to the system's default encoding
    int i;
    int length = [self cStringLength];
    unsigned char *bytes = [self __compact8BitBytes];
    unsigned char *chars = MallocAtomic(sizeof(unichar)*(length+1));

    for (i = 0; i < length; i++) {
	register unsigned char c = bytes[i];
	chars[i] = islower(c) ? toupper(c) : c;
    }
    
    chars[i] = 0;

    return AUTORELEASE([[NSOwned8BitString alloc]
                           initWithCString:chars length:length copy:NO]);
}

/* Working with C strings */

- (void)getCString:(char*)buffer maxLength:(unsigned int)maxLength
  range:(NSRange)aRange remainingRange:(NSRange*)leftoverRange
{
    unsigned char* bytes = [self __compact8BitBytes];
    unsigned int toMove = MIN(maxLength, aRange.length);
    unsigned int cLength = [self cStringLength];
    
    if (aRange.location + aRange.length > cLength) {
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, cLength]);
    }

    if (leftoverRange) {
	leftoverRange->location = aRange.location + toMove;
	leftoverRange->length = cLength - leftoverRange->location;
    }
    memcpy(buffer, bytes + aRange.location, toMove);
    if (toMove < maxLength)
	buffer[toMove] = '\0';
}

- (BOOL)writeToFile:(NSString *)path atomically:(BOOL)flag
{
    // UNICODE - remove this
    NSData *data;
    data = [self dataUsingEncoding:[NSString defaultCStringEncoding]];
    return writeToFile(path, data, flag);
}

- (Class)classForCoder
{
    return [NS8BitString class];
}

- (void)encodeWithCoder:(NSCoder*)aCoder
{
    const unsigned char* bytes = [self __compact8BitBytes];
    int length = [self cStringLength];
    
    [aCoder encodeValueOfObjCType:@encode(int) at:&length];
    [aCoder encodeArrayOfObjCType:@encode(char) count:length at:bytes];
}

- (id)initWithCoder:(NSCoder*)aDecoder
{
    unsigned char *bytes;
    int length;

    RELEASE(self); self = nil;

    [aDecoder decodeValueOfObjCType:@encode(int) at:&length];
    bytes = MallocAtomic (length + 1);
    [aDecoder decodeArrayOfObjCType:@encode(char) count:length at:bytes];
    bytes[length] = '\0';
    return [[NSOwned8BitString alloc] initWithCString:bytes length:length copy:NO];
}

- (NSString *)stringRepresentation
{
    const unsigned char *cString;
    int i, length;

    cString = [self __compact8BitBytes];
    length = [self cStringLength];
    
    if (cString == NULL)    return @"\"\"";
    if (length  == 0)       return @"\"\"";
    if (cString[0] == '\0') return @"\"\"";
    
    /* Check if the string can be parsed as a STRING token by the property list
       parser. Otherwise we must enclose it in double quotes. */
    if (lf_isPlistBreakChar(cString[0])) {
        return lf_quoteString(cString, length);
    }

    for(i = 1; i < length; i++) {
        if (lf_isPlistBreakChar(cString[i]))
	    return lf_quoteString(cString, length);
    }

    return self;
}

- (id)copyWithZone:(NSZone*)zone
{
    Class clazz;
    int length;
    
    if (NSShouldRetainWithZone(self, zone))
	return RETAIN(self);

    length = [self cStringLength];
    
    clazz = length < 255
        ? [NSShortInline8BitString class]
        : [NSInline8BitString class];
    
    return [[clazz allocForCapacity:length zone:zone]
                   initWithCString:[self __compact8BitBytes] length:length];
}

- (NSData *)dataUsingEncoding:(NSStringEncoding)encoding
  allowLossyConversion:(BOOL)flag;
{
    /* NS8BitString */
    if (encoding == [NSString defaultCStringEncoding]) {
        unsigned      len;
        unsigned char *buf = NULL;
        
        len  = [self cStringLength];
        buf = NSZoneMalloc(NULL, sizeof(unsigned char) * len + 1);
        [self getCString:buf];
        buf[len] = '\0';
        return [NSData dataWithBytesNoCopy:buf length:strlen(buf)];
    }
    else if (encoding == NSASCIIStringEncoding) {
        register unsigned len = [self cStringLength];
        register unsigned i;
        register unsigned char *buf;
            
        buf = NSZoneMalloc(NULL, sizeof(char) * len + 1);
        buf[len] = '\0';
            
        [self getCString:buf];            
        if (!flag) {
            /* check for strict ASCII */
            for (i = 0; i < len; i++)
                if (buf[i] > 127) return nil;
        }
        return [NSData dataWithBytesNoCopy:buf length:len];
    }
    else
        return [super dataUsingEncoding:encoding allowLossyConversion:flag];
}

- (id)mutableCopyWithZone:(NSZone*)zone
{
    return [[NSMutableSimple8BitString allocWithZone:zone]
	initWithCString:[self __compact8BitBytes]
	length:[self cStringLength] copy:YES];
}

@end /* NS8BitString */

/*
 * Null terminated CString containing characters inline
 */
 
@implementation NSInline8BitString

+ (id)allocForCapacity:(unsigned int)capacity zone:(NSZone *)zone
{
    NSInline8BitString *str = NSAllocateObject(self, capacity, zone);
    str->cLength = -1;
    return str;
}

- (id)init
{
    if (cLength != -1) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    self->cLength = 0;
    self->cString[0] = 0;
    return self;
}

- (id)initWithCString:(const char*)byteString length:(unsigned int)length
{
    if (cLength != -1)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    self->cLength = length;
    memcpy(self->cString, byteString, length);
    self->cString[length] = 0;
    return self;
}
- (id)initWithCharacters:(const unichar *)chars length:(unsigned int)length
{
    /* it must be ensured that char values are below 256 by the cluster ! */
    register unsigned i;
    if (cLength != -1) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    
    for (i = 0; i < length; i++)
        self->cString[i] = chars[i];
    self->cString[i] = '\0';
    self->cLength = i;
    return self;
}

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NSInline8BitString_dealloc_count++;
    NSInline8BitString_total_len += self->cLength == -1 ? 0 : self->cLength;
    [super dealloc];
}
#endif

- (const char *)cString
{
    return self->cString;
}

- (unsigned int)cStringLength
{
    return (self->cLength == -1) ? 0 : self->cLength;
}

- (unsigned int)length
{
    return (self->cLength == -1) ? 0 : self->cLength;
}

- (unichar)characterAtIndex:(unsigned int)index
{
    if (self->cLength == -1 || index >= self->cLength) {
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"index %d out of range in string %x of length %d",
	    	index, self, self->cLength]);
    }
    // ENCODING
    return self->cString[index];
}

- (NSString *)substringWithRange:(NSRange)aRange
{
    if (aRange.location + aRange.length > self->cLength)
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, cLength]);

    if (aRange.length == 0)
	return @"";

    return AUTORELEASE([[NSRange8BitString alloc]
                           initWithString:self
                           bytes:(self->cString + aRange.location)
                           length:aRange.length]);
}

- (char *)__compact8BitBytes
{
    return self->cString;
}

@end /* NSInline8BitString */

@implementation NSShortInline8BitString

+ (id)allocForCapacity:(unsigned int)capacity zone:(NSZone*)zone
{
    NSShortInline8BitString *str = NSAllocateObject(self, capacity, zone);
    str->cLength = 255;
    return str;
}

- (id)init
{
    if (cLength != 255) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    self->cLength = 0;
    self->cString[0] = 0;
    return self;
}

- (id)initWithCString:(const char*)byteString length:(unsigned int)length
{
    if (cLength != 255) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    self->cLength = length;
    memcpy(self->cString, byteString, length);
    self->cString[length] = 0;
    return self;
}
- (id)initWithCharacters:(const unichar *)chars length:(unsigned int)length
{
    /* it must be ensured that char values are below 256 by the cluster ! */
    register unsigned i;
    if (cLength != 255) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    
    for (i = 0; i < length; i++)
        self->cString[i] = chars[i];
    self->cString[i] = '\0';
    self->cLength = i;
    return self;
}

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NSShortInline8BitString_dealloc_count++;
    NSShortInline8BitString_total_len +=
        self->cLength == 255 ? 0 : self->cLength;
    [super dealloc];
}
#endif

- (const char *)cString
{
    return self->cString;
}

- (unsigned int)cStringLength
{
    return (self->cLength == 255) ? 0 : self->cLength;
}

- (unsigned int)length
{
    return (self->cLength == 255) ? 0 : self->cLength;
}

- (unichar)characterAtIndex:(unsigned int)index
{
    if (self->cLength == 255 || index >= self->cLength) {
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"index %d out of range in string %x of length %d",
	    	index, self, self->cLength]);
    }
    // ENCODING
    return self->cString[index];
}

- (NSString *)substringWithRange:(NSRange)aRange
{
    if (aRange.location + aRange.length > self->cLength)
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"range (%d,%d) in string %x of length %d",
	    	aRange.location, aRange.length, self, cLength]);

    if (aRange.length == 0)
	return @"";

    return AUTORELEASE([[NSRange8BitString alloc]
                           initWithString:self
                           bytes:(self->cString + aRange.location)
                           length:aRange.length]);
}

- (char *)__compact8BitBytes
{
    return self->cString;
}

@end /* NSShortInline8BitString */

@implementation NSCharacter8BitString

- (id)init
{
    self->c[0] = '\0';
    self->c[1] = '\0';
    return self;
}

- (id)initWithCString:(const char*)byteString length:(unsigned int)length
{
    if (length != 1) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"%@ can only handle a singe char", [self class]]);
    }
    self->c[0] = byteString[0];
    self->c[1] = '\0';
    
    return self;
}
- (id)initWithCharacters:(const unichar *)chars length:(unsigned int)length
{
    if (length != 1) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"%@ can only handle a singe char", [self class]]);
    }
    self->c[0] = (char)chars[0];
    self->c[1] = '\0';
    
    return self;
}

- (const char *)cString
{
    return &(self->c[0]);
}

- (unsigned int)cStringLength
{
  return 1;
}

- (unsigned int)length
{
  return 1;
}

- (unichar)characterAtIndex:(unsigned int)index
{
    if (index != 0) {
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"index %d out of range in string %x of length 1",
	    	index, self]);
    }
    // ENCODING
    return self->c[0];
}

- (NSString *)substringWithRange:(NSRange)aRange
{
  if (aRange.location == 0 && aRange.length == 1)
    return self;
  if (aRange.length == 0)
    return @"";

  THROW([[IndexOutOfRangeException alloc] 
	  initWithFormat:@"range (%d,%d) in string %x of length 1",
	  aRange.location, aRange.length, self]);
  return nil;
}

- (char *)__compact8BitBytes
{
    return &(self->c[0]);
}

@end /* NSCharacter8BitString */


/*
 * String containing non-owned, zero termintated c-string
 */

@implementation NSNonOwned8BitString

- (id)init
{
    if (self->cLength || self->cString)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    self->cLength = 0;
    self->cString = "";
    return self;
}

- (id)initWithCString:(char *)byteString
  length:(unsigned int)length copy:(BOOL)flag
{
    if (self->cLength || self->cString) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }
    if (flag) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s with flag set to YES to %@ to instance", 
	    sel_get_name(_cmd), NSStringFromClass([self class])]);
    }
    cLength = length;
    cString = byteString;
    return self;
}

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NSNonOwned8BitString_dealloc_count++;
    NSNonOwned8BitString_total_len += self->cLength;
    [super dealloc];
}
#endif

- (const char *)cString
{
    return self->cString;
}

- (unsigned int)cStringLength
{
    return self->cLength;
}

- (unsigned int)length
{
    return self->cLength;
}

- (unichar)characterAtIndex:(unsigned int)index
{
    if (self->cLength == -1 || index >= cLength)
	THROW([[IndexOutOfRangeException alloc] 
	    initWithFormat:@"index %d out of range in string %x of length %d",
	    	index, self, cLength]);
    // ENCODING
    return self->cString[index];
}

- (NSString *)substringWithRange:(NSRange)aRange
{
    if (aRange.location + aRange.length > cLength)
	THROW([[IndexOutOfRangeException alloc] 
                  initWithFormat:@"range (%d,%d) in string %x of length %d",
                  aRange.location, aRange.length, self, cLength]);

    if (aRange.length == 0)
	return @"";

    return AUTORELEASE([[NSRange8BitString alloc]
                           initWithString:self
                           bytes:cString + aRange.location
                           length:aRange.length]);
}

- (char*)__compact8BitBytes
{
    return self->cString;
}

@end /* NSNonOwned8BitString */

@implementation NSOwned8BitString

- (id)init
{
    if (self->cLength || self->cString)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);

    self->cLength = 0;
    self->cString = NSZoneMallocAtomic([self zone], sizeof(char));
    self->cString[0] = 0;
    return self;
}

- (id)initWithCString:(char*)byteString
  length:(unsigned int)length
  copy:(BOOL)flag
{
    if (self->cLength || self->cString)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);

    self->cLength = length;
    if (flag) {
	self->cString = NSZoneMallocAtomic([self zone], sizeof(char)*(length+1));
	memcpy(self->cString, byteString, length);
	self->cString[self->cLength] = 0;
    }
    else
	self->cString = byteString;
    return self;
}

- (void)dealloc
{
#if COLLECT_STRING_CLUSTER_STATISTICS
    NSOwned8BitString_dealloc_count++;
    NSOwned8BitString_total_len += self->cLength;
#endif
    lfFree(self->cString);
    [super dealloc];
}

@end /* NSOwned8BitString */

#if 1
@implementation NXConstantString

- (oneway void)release
{
}

- (id)retain
{
    return self;
}

- (unsigned)retainCount
{
    return 1;
}

- (id)autorelease
{
    return self;
}

- (void)dealloc
{
    [self shouldNotImplement:_cmd];
}

@end /* NXConstantString */
#else
@interface DummyNXConstantString : NSNonOwned8BitString
@end

@implementation DummyNXConstantString
- (oneway void)release
{
}

- (id)retain
{
    return self;
}

- (unsigned)retainCount
{
    return 1;
}

- (id)autorelease
{
    return self;
}

- (void)dealloc
{
    [self shouldNotImplement:_cmd];
}

+ (void)load
{
    static BOOL didLoad = NO;

    if (!didLoad) {
        Class metaClass;
        Class constantStringClass;
        Class constantStringMetaClass;

        didLoad = YES;

        metaClass               = ((DummyNXConstantString*)self)->isa;
        constantStringClass     = objc_lookup_class ("NXConstantString");
        constantStringMetaClass = constantStringClass->class_pointer;
        
        memcpy(constantStringClass,     self,      sizeof (struct objc_class));
        memcpy(constantStringMetaClass, metaClass, sizeof (struct objc_class));
        
        constantStringClass->name
            = constantStringMetaClass->name
            = "NXConstantString";

        class_add_behavior(constantStringClass, self);
    }
}

@end /* DummyNXConstantString */
#endif

@implementation NSNonOwnedOpen8BitString

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NSNonOwnedOpen8BitString_dealloc_count++;
    NSNonOwnedOpen8BitString_total_len += self->cLength;
    [super dealloc];
}
#endif

- (const char*)cString
{
    unsigned char *str;
    
    str = NSZoneMallocAtomic([self zone], sizeof(char)*(self->cLength + 1));
    memcpy(str, self->cString, self->cLength);
    str[cLength] = 0;
#if !LIB_FOUNDATION_BOEHM_GC
    [NSAutoreleasedPointer autoreleasePointer:str];
#endif
    return str;
}

@end /* NSNonOwnedOpen8BitString */

@implementation NSOwnedOpen8BitString

#if COLLECT_STRING_CLUSTER_STATISTICS
- (void)dealloc
{
    NSOwnedOpen8BitString_dealloc_count++;
    NSOwnedOpen8BitString_total_len += self->cLength;
    [super dealloc];
}
#endif

- (id)initWithCString:(char *)byteString
  length:(unsigned int)length copy:(BOOL)flag
{
    if (self->cLength || self->cString) {
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);
    }

    self->cLength = length;
    if (flag) {
	cString = NSZoneMallocAtomic([self zone], sizeof(char)*(length));
	memcpy(cString, byteString, length);
    }
    else
	cString = byteString;
    return self;
}

- (const char *)cString
{
    unsigned char *str = MallocAtomic(sizeof(char)*(self->cLength + 1));
    
    memcpy(str, self->cString, self->cLength);
    str[self->cLength] = 0;
#if !LIB_FOUNDATION_BOEHM_GC
    [NSAutoreleasedPointer autoreleasePointer:str];
#endif
    return str;
}

@end /* NSOwnedOpen8BitString */

@implementation NSRange8BitString

- (id)initWithString:(NSString *)aParent 
  bytes:(char *)bytes length:(unsigned int)length
{
    if (self->cLength || self->cString)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot send %s to non-mutable instance", sel_get_name(_cmd)]);

    self->cString = bytes;
    self->cLength = length;
    self->parent  = RETAIN(aParent);
    return self;
}

- (void)dealloc
{
#if COLLECT_STRING_CLUSTER_STATISTICS
    NSRange8BitString_dealloc_count++;
    NSRange8BitString_total_len += self->cLength;
#endif
    RELEASE(self->parent);
    [super dealloc];
}

- (NSString *)substringWithRange:(NSRange)aRange
{
    if (aRange.location + aRange.length > cLength) {
	THROW([[IndexOutOfRangeException alloc] 
                  initWithFormat:@"range (%d,%d) in string %x of length %d",
                  aRange.location, aRange.length, self, cLength]);
    }
    
    if (aRange.length == 0)
	return @"";

    return AUTORELEASE([[NSRange8BitString alloc] 
                           initWithString:parent
                           bytes:cString+aRange.location 
                           length:aRange.length]);
}

@end /* NSRange8BitString */

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
